import { globalLookup } from "@/hook/global";
import { colorPicker, contextmenu, groupcontextmenu, handleGroupContextmenuFromSignalItem } from "./handle-contextmenu";
import { reactive, ref } from "vue";
import { findViewIndexByLink } from "@/hook/wave-container-view";
import { saveViewApi } from "@/api";

export const groupColors = [
    {
        color: '#dadce0',
        used: false
    },
    {
        color: '#7ca0d9',
        used: false
    },
    {
        color: '#e3847d',
        used: false
    },
    {
        color: '#fdd663',
        used: false
    },
    {
        color: '#7bbe8f',
        used: false
    },
    {
        color: '#ff8bcb',
        used: false
    },
    {
        color: '#c58af9',
        used: false
    },
    {
        color: '#73ccdf',
        used: false
    },
    {
        color: '#fcad70',
        used: false
    }
];

const color2Index = {};
for (let i = 0; i < groupColors.length; ++ i) {
    const color = groupColors[i].color;
    color2Index[color] = i;
}

/**
 * @description 当前的组
 * @type {WaveRenderSidebarItem[]}
 */
export const currentGroups = reactive([]);

export async function updateCurrentGroups() {
    currentGroups.length = 0;
    for (const view of globalLookup.currentWiresRenderView) {
        if (view.renderType === 1) {
            // 1 代表当前为 group
            currentGroups.push(view);      
        }
    }
}

export const groupColorDispatcher = {
    /**
     * @description 获取颜色
     * @returns {string}
     */
    get() {
        let color = groupColors[0];
        for (const c of groupColors) {
            if (!c.used) {
                c.used = true;
                return c.color;
            }
        }
        return color;
    },
    /**
     * @description 归还颜色
     * @param {string} color 
     */
    put(color) {
        const index = color2Index[color];
        if (index !== undefined && index < groupColors.length) {
            groupColors[index].used = false;
        }
    }
};

export function makeColorStyle(color) {
    return `background-color: ${color};` 
}

// 创建一个组，并且从 group 颜色调度器拿走一个颜色
export function createGroup() {
    const signal = contextmenu.currentWire;
    const wireIndex = contextmenu.currentWireIndex;

    if (signal) {
        const renderView = globalLookup.currentWiresRenderView;
        if (wireIndex < renderView.length) {
            const view = renderView[wireIndex];
            view.renderType = 1;
            const groupLink = view.signalInfo.link + '-group';
            view.signalInfo.link = groupLink;
            view.groupInfo = {
                name: '',
                color: groupColorDispatcher.get(),
                collapse: false
            };
            view.children.push({
                signalInfo: {
                    name: signal.name,
                    link: signal.link,
                    size: signal.size,
                    parentLink: groupLink
                },
                groupInfo: {
                    name: '',
                    color: '',
                    collapse: false
                },
                renderType: 0,
                children: []
            });

            handleGroupContextmenuFromSignalItem();
        }
    }
    contextmenu.show = false;
    globalLookup.render();
    updateCurrentGroups();

    // 保存视图
    saveViewApi({ views: true });
}

export function cancelGroup() {
    const currentGroupView = groupcontextmenu.currentGroupView;
    if (currentGroupView) {
        const link = currentGroupView.signalInfo.link;
        const renderView = globalLookup.currentWiresRenderView;

        const i = findViewIndexByLink(link);
        if (i > 0) {
            // 此时 renderView[i] 就是 currentGroupView，我们需要替换它
            // 把 i 之后的元素保存起来
            const tailElements = renderView.slice(i + 1);
            renderView.length = i;
            // 把 i 的 children 先加进来，然后再把保存的元素加进来
            currentGroupView.children.forEach(view => renderView.push(view));
            tailElements.forEach(view => renderView.push(view));
            // 归换颜色
            groupColorDispatcher.put(currentGroupView.groupInfo.color);
        }
    }
    
    groupcontextmenu.show = false;
    updateCurrentGroups();

    // 保存视图
    saveViewApi({ views: true });
}


export function deleteGroup() {
    const currentGroupView = groupcontextmenu.currentGroupView;
    if (currentGroupView) {
        const link = currentGroupView.signalInfo.link;
        const renderView = globalLookup.currentWiresRenderView;

        const i = findViewIndexByLink(link);
        if (i > 0) {
            const tailElements = renderView.slice(i + 1);
            renderView.length = i;
            tailElements.forEach(view => renderView.push(view));

            currentGroupView.children.forEach(view => {
                const signal = globalLookup.link2CurrentWires.get(view.signalInfo.link);
                if (signal) {
                    globalLookup.currentWires.delete(signal);
                    globalLookup.link2CurrentWires.delete(signal.link);
                    delete globalLookup.currentSignalValues[signal.link];
                }
            });
            groupColorDispatcher.put(currentGroupView.groupInfo.color);
        }
    }

    groupcontextmenu.show = false;
    updateCurrentGroups();

    // 保存视图
    saveViewApi({ waves: true });
}