import { controller } from "@/components/treeview/signals";
import { globalLookup } from "./global";
import { WaveContainerView } from "./wave-container-view";
import { isScope, isVariable } from "./utils";
import { BSON } from "bson";
import { Id2Pivot, orderedTimes, UserPivots } from "@/components/pivot/pivot-view";
import { TreeviewExpandSignals } from "@/components/treeview/modules";
import { controlPanel } from "@/components/right-nav";
import { saveViewApi } from "@/api";

export const recoverConfig = {
    /**
     * @type {Map<string, IRenderOption>} 其中的 string 是形如 `cpu.alu.add1` 这样的字符串
     */
    waves: new Map(),

    /**
     * @type {WaveRenderSidebarItem[] | undefined}
     */
    views: undefined,

    /**
     * @type {Pstate | undefined}
     */
    state: undefined,

    /**
     * @type {import("@/components/pivot/pivot-view").UserPivot[] | undefined}
     */
    pivots: undefined,

    /**
     * @type {Set<string> | undefined} nameClass 的集合，代表所有需要打开的 treeview 中的 scope
     */
    treeviewExpands: undefined,

    /**
     * @type {number} 右侧的面版的打开的编号，-1 代表没有面板被打开
     */
    rightNavIndex: -1
};

async function findRecoverFile(recoverJsonPath) {
    try {
        const response = await fetch(recoverJsonPath);        
        if (!response.ok) {
            return undefined;
        }
        const recoverBuffer = await response.arrayBuffer();
        const recoverJson = BSON.deserialize(recoverBuffer);        
        return recoverJson;
    } catch (error) {
        console.log('[findRecoverFile] error: ' + error);
        return undefined;
    }
}

/**
 * 
 * @param {string} inputFile 读取的 vcd 文件的路径
 * @param {string} inputViewFile 读取的 vcd view 文件的路径，它如果不存在则自动创建
 */
async function attemptRecover(inputFile, inputViewFile) {
    let recoverResult = await findRecoverFile(inputViewFile);
    if (recoverResult !== undefined) {
        globalLookup.originVcdViewFile = inputViewFile;
        return recoverResult;
    }

    // 不存在则创建一个和 inputFile 同名的 view 文件
    if (inputFile.endsWith('.vcd')) {
        inputFile = inputFile.slice(0, -4);
    }
    globalLookup.originVcdViewFile = inputFile + '.view';
    return await findRecoverFile(globalLookup.originVcdViewFile);
}

/**
 * 
 * @param {string} inputFile 读取的 vcd 文件的路径
 * @param {string} inputViewFile 读取的 vcd view 文件的路径，它如果不存在则自动创建
 */
export async function recoverFromInputFile(inputFile, inputViewFile) {
    globalLookup.originVcdFile = inputFile;
    // 先尝试从 inputViewFile 中寻找，如果找不到，则从同名的 .view 中寻找
    const recoverJson = await attemptRecover(inputFile, inputViewFile);
    recoverFromJson(recoverJson);
}

export function recoverFromJson(recoverJson) {
    if (recoverJson === undefined) {
        return;
    }

    // 加载 waves
    const waves = recoverJson.waves;
    if (!recoverConfig.waves) {
        recoverConfig.waves = new Map();
    }
    if (waves instanceof Array && waves.length > 0) {
        for (const wave of waves) {
            const name = wave.name;
            const option = wave.option;
            if (name === undefined || option === undefined || recoverConfig.waves.has(name)) {
                continue;
            }
            recoverConfig.waves.set(name, option);
        }
    }

    // 加载 views
    const views = recoverJson.views;
    if (views instanceof Array && views.length > 0) {
        recoverConfig.views = recoverJson.views;
    }

    // 加载 state
    recoverConfig.state = recoverJson.state;

    // 加载 pivots
    const pivots = recoverJson.pivots;
    if (pivots instanceof Array && pivots.length > 0) {
        recoverConfig.pivots = pivots;
    }

    // 加载 treeviewExpands
    const treeviewExpands = recoverJson.treeviewExpands;
    if (treeviewExpands instanceof Array && treeviewExpands.length > 0) {
        recoverConfig.treeviewExpands = treeviewExpands;
    }

    // 加载 rightNavIndex
    const rightNavIndex = recoverJson.rightNavIndex;
    if (rightNavIndex !== undefined) {
        recoverConfig.rightNavIndex = rightNavIndex;
    }   
}

class PrefixNode {
    constructor() {
        this.children = new Map();
        this.payload = undefined;
    }
}



/**
 * @description 恢复现场的函数，主要恢复两个变量 currentWiresRenderView 和 currentSignalRenderOptions
 * @param {ScopeItem[]} topModules
 */
export function recoverSession(topModules) {
    // 恢复 waves
    // 利用前缀树 记忆化搜索
    const waves = recoverConfig.waves;
    for (const [name, option] of waves.entries()) {
        const result = names2wire(topModules, name);
        if (result) {
            // 找到了 进行渲染设置
            WaveContainerView.add(result);
            globalLookup.currentSignalRenderOptions.set(result.link, option);            
            controller.lastSignal = result;
        } else {
            console.log('fail to recover ' + name);
        }
    }

    // 恢复 views
    // 如果存在 views
    if (recoverConfig.views !== undefined) {
        const renderViews = globalLookup.currentWiresRenderView;
        renderViews.length = 0;
        for (const view of recoverConfig.views) {
            // 恢复 names -> link
            const signalNames = view.signalInfo.link;
            if (signalNames.includes('.')) {
                const signal = names2wire(topModules, signalNames);
                view.signalInfo.link = signal.link;
            }

            const parentLink = view.signalInfo.parentLink;
            if (parentLink && parentLink.endsWith('-group')) {
                const parentNames = parentLink.slice(0, -6);
                const parentSignal = names2wire(topModules, parentNames);
                view.signalInfo.parentLink = parentSignal.link + '-group';
            }

            for (const child of view.children) {
                const signalNames = child.signalInfo.link;
                if (signalNames.includes('.')) {
                    child.signalInfo.link = names2wire(topModules, signalNames).link;
                }
            }

            renderViews.push(view);
        }
    }

    // 恢复 state
    const pstate = recoverConfig.state;
    if (pstate !== undefined) {
        Object.assign(globalLookup.pstate, pstate);
    }

    // 恢复 pivots
    const pivots = recoverConfig.pivots;
    if (pivots !== undefined && pivots instanceof Array) {
        for (const pivot of pivots) {
            const time = pivot.time;
            const id = pivot.id;            
            UserPivots.set(time, pivot);
            Id2Pivot.set(id, pivot);
            orderedTimes.push(time);
        }
        orderedTimes.sort((a, b) => a - b);
    }

    // 恢复 treeviewExpands
    const treeviewExpands = recoverConfig.treeviewExpands;
    if (treeviewExpands !== undefined && treeviewExpands instanceof Array) {
        for (const nameClass of treeviewExpands) {
            const scope = names2wire(topModules, nameClass);
            if (scope !== undefined) {
                TreeviewExpandSignals.add(scope);
            } else {
                console.log('fail to recover [treeviewExpands] item: ' + nameClass);
            }
        }
    }

    // 恢复 rightNavIndex
    const rightNavIndex = recoverConfig.rightNavIndex;
    if (rightNavIndex !== undefined) {
        controlPanel.currentIndex = rightNavIndex;
    }

    globalLookup.render();

    recoverConfig.pivots = null;
    recoverConfig.state = null;
    recoverConfig.views = null;
    recoverConfig.treeviewExpands = null;
    recoverConfig.waves = null;
    recoverConfig.rightNavIndex = null;

    // 更新后端缓存
    saveViewApi({all: true});
}

/**
 * 
 * @param {ScopeItem[]} topModules 
 * @param {string} name
 * @returns {WireItem | undefined} 
 */
function names2wire(topModules, name) {    
    const deps = name.split('.');
    let layer = 0;
    let nodes = topModules;
    let result = undefined;
    let lastNode = undefined;

    while (layer < deps.length) {
        // 从当前 nodes 中匹配 className
        const targetName = deps[layer];
        let targetNode = nodes.filter(n => n.name == targetName)[0];
        if (targetNode === undefined) {
            break;
        }

        // 搜索完成直接返回
        if (layer === deps.length - 1) {
            result = targetNode;
            if (result.parent === undefined) {
                result.parent = lastNode;
            }
            break;
        }

        if (isScope(targetNode)) {
            nodes = targetNode.body;
        } else {
            break;
        }
        lastNode = targetNode;
        layer ++;
    }

    if (result !== undefined) {
        return result;
    }
    return undefined;
}

/**
 * 
 * @param {string} link 
 * @returns {string}
 */
function link2names(link) {
    const names = [];
    const signal = globalLookup.link2CurrentWires.get(link);
    if (signal) {
        if (signal.nameClass !== undefined) {
            return signal.nameClass;
        }

        let node = signal;
        while (node !== undefined) {
            names.push(node.name);
            node = node.parent;
        }
    }

    return names.reverse().join('.');
}

/**
 * @template T
 * @param {T} obj 
 * @returns {T}
 */
function stableClone(obj) {
    if (window.structuredClone) {
        return structuredClone(obj);
    }
    return JSON.parse(JSON.stringify(obj));
}


/**
 * @typedef {Object} SavePayloadConfig
 * @property {boolean} waves
 * @property {boolean} views
 * @property {boolean} state
 * @property {boolean} pivots
 * @property {boolean} treeviewExpands
 * @property {boolean} rightNavIndex
 * @property {boolean} all
 */



/**
 * @param {SavePayloadConfig} config 
 * @param {keyof SavePayloadConfig} property 
 * @returns 
 */
function defaultFalseWrapper(config, property) {
    if (config === undefined) {
        return false;
    }

    if (config[property] === undefined) {
        return false;
    }

    return Boolean(config[property]);
}

/**
 * 
 * @param {SavePayloadConfig | undefined} config 
 * @returns 
 */
export function makeSaveViewPayload(config) {
    const payload = {};
    config = config || {};

    const all = defaultFalseWrapper(config, 'all');

    // 波形 profile
    if (all || defaultFalseWrapper(config, 'waves')) {
        // 有 waves 也一定要有 views
        config.views = true;

        const waves = [];
        for (const wire of globalLookup.currentWires) {
            const option = globalLookup.currentSignalRenderOptions.get(wire.link) || {};        
            const name = link2names(wire.link);
            option.highlight = 0;
            waves.push({ name, option });
        }
        Object.assign(payload, { waves });
    }


    // 视图列表
    if (all || defaultFalseWrapper(config, 'views')) {
        const views = [];
        for (const view of globalLookup.currentWiresRenderView) {
            // link 转换成 names
            const saveView = stableClone(view);
            const viewLink = view.signalInfo.link;
            saveView.signalInfo.link = link2names(viewLink);
            const parentLink = saveView.signalInfo.parentLink;
            if (parentLink !== undefined) {
                saveView.signalInfo.parentLink = parentLink.endsWith('-group') ? link2names(parentLink.slice(0, -6)) + '-group' : viewLink;
            }
            // children 只会有一层
            for (const child of saveView.children) {
                child.signalInfo.link = link2names(child.signalInfo.link);
            }
            views.push(saveView);
        }
        Object.assign(payload, { views });
    }


    // 窗口状态
    if (all || defaultFalseWrapper(config, 'state')) {
        // proxy 对象在一定情况下无法序列化
        const state = JSON.parse(JSON.stringify(globalLookup.pstate));
        Object.assign(payload, { state });
    }


    // 信标相关
    if (all || defaultFalseWrapper(config, 'pivots')) {
        const pivots = [];
        for (const pivot of UserPivots.values()) {
            pivots.push(pivot);
        }
        Object.assign(payload, { pivots });
    }


    // 右侧信号展开信息
    if (all || defaultFalseWrapper(config, 'treeviewExpands')) {
        const treeviewExpands = [];
        for (const signal of TreeviewExpandSignals) {
            treeviewExpands.push(signal.nameClass);
        }
        Object.assign(payload, { treeviewExpands });
    }


    // 右侧控制面板打开序号
    if (all || defaultFalseWrapper(config, 'rightNavIndex')) {
        const rightNavIndex = controlPanel.currentIndex;
        Object.assign(payload, { rightNavIndex });
    }

    return payload;
}