import { Stream } from "stream";
import { globalLookup } from "./global";

/**
 * 
 * @returns {Promise<Stream>}
 */
export async function getVcdStream() {
    const ostream = await window.getVcdStream();
    return ostream;
}


export function debounceWrapper(fn, delay) {
    let timer
    return function (...params) {
        if (timer) {
            clearTimeout(timer)
        }
        timer = setTimeout(() => {
            fn(...params);
        }, delay)
    }
}


export function makeIconClass(mod) {
    if (mod === undefined) {
        return '';
    }
    switch (mod.type) {
        case 'module':      return 'icon-memory-chip';
        case 'begin':       return 'icon-brackets';
        case 'fork':        return 'icon-fork';
        case 'function':    return 'icon-function';
        case 'task':        return 'icon-task';

        case 'event':       return 'icon-prevent';
        case 'integer':     return 'icon-integer';
        case 'parameter':   return 'icon-parameter';
        case 'real':        return 'icon-real';
        case 'realtime':    return 'icon-wave-square';
        case 'reg':         return 'icon-register';
        case 'supply0':     return 'icon-wave-square';
        case 'supply1':     return 'icon-wave-square';
        case 'time':        return 'icon-wave-square';
        case 'tri':         return 'icon-wave-square';
        case 'triand':      return 'icon-wave-square';
        case 'trior':       return 'icon-wave-square';
        case 'trireg':      return 'icon-wave-square';
        case 'tri0':        return 'icon-wave-square';
        case 'tri1':        return 'icon-wave-square';
        case 'wand':        return 'icon-wave-square';
        case 'wire':        return 'icon-wave-square';
        case 'wor':         return 'icon-wave-square';
        case 'string':      return 'icon-string';
        default:
            break;
    }
}


/**
 * @param { string } searchString
 * @param {{
 * kind: 'var' | 'scope',
 * type: 'wire' | 'reg' | 'module',
 * name: string,
 * size?: number,
 * link?: string
 * }} module 
 * @param { Set<string> } searchScope
 * @param { boolean } caseSensitivity
 * @param { boolean } displayParentOnly
 * @returns { { htmlString: string, module } | null }
*/
export function makeSearchResultItem(searchString, module, searchScope, caseSensitivity, displayParentOnly) {
    if (searchScope.has(module.type)) {
        let pattern = module.name;
        if (!caseSensitivity) {
            pattern = pattern.toLowerCase();
            searchString = searchString.toLowerCase();
        }
        if (pattern.includes(searchString)) {
            let p = module;

            const deps = [];
            while (p) {
                if (p.name && p.type) {
                    deps.push(p);
                }
                p = p.parent;
                if (displayParentOnly && deps.length == 2) {
                    break;
                }
            }

            let htmlString = '';
            let currentCounts = 0;
            for (let i = deps.length - 1; i >= 0; -- i) {
                const mod = deps[i];
                // const displayName = mod.name.length > 6 ? mod.name.substring(0, 6) + '...' : mod.name;
                const iconClass = makeIconClass(mod);
                const iconText = `<span class="iconfont ${iconClass}"></span>&ensp;${mod.name}`;
                htmlString += iconText;

                if (i > 0) {
                    htmlString += '<div class="dep-arrow"></div>';
                }
            }

            return {
                htmlString,
                module
            };
        } 
    }
    return null;
}

/**
 * 
 * @param {'bit' | 'vec'} kind
 * @param {BigInt} value 
 * @param {BigInt} mask 
 * @returns {BigInt | string}
 */
export function getWireValueCaption(kind, value, mask) {
    if (kind === 'bit') {
        if (value === 2) {
            return 'x';
        }
        return value;
    } else if (kind === 'vec') {
        if (!mask) {
            return value;
        }
        // mask 不为空，代表存在问题
        if (value) {
            return '?';
        } else {
            return 'x';
        }
    }
}

/**
 * 
 * @param {changoItem} changoItem 
 * @param {number} time 
 * @returns {BigInt | string}
 */
export function getSmartCurrentSignalValue(changoItem, time) {
    const { wave, kind } = changoItem;
    if (wave === undefined || wave.length === 0) {
        return 'x';        
    } else if (wave.length === 1) {
        return getWireValueCaption(kind, wave[0][1], wave[0][2]);
    } else {
        return findCurrentSignalValue(kind, wave, time);   
    }
}

export async function updateWireCurrentValue(time) {
    const chango = globalLookup.chango;
    if (time === undefined) {
        time = globalLookup.currentTime;
    }

    const gcd = globalLookup.tgcd || 1;
    time = time / gcd;
    
    const currentSignalValues = globalLookup.currentSignalValues;

    for (const signal of globalLookup.currentWires) {
        const changoItem = chango[signal.link];
        currentSignalValues[signal.link] = getSmartCurrentSignalValue(changoItem, time);
    }
}

/**
 * @param {'bit' | 'vec'} kind
 * @param {Array<Array<string | number>>} wave 
 * @param {number} time 
 */
export function findCurrentSignalValue(kind, wave, time) {
    const times = wave.map(p => p[0]);

    // 二分查找，并将结果存入 i
    let i = 0, j = wave.length - 1;
    while (i < j) {
        if (times[i] === time) {
            break;
        }
        if (times[j] <= time) {
            i = j;
            break;
        }
        if (j - i === 1) {
            break;
        }
        const mid = (i + j) >> 1;
        if (times[mid] > time) {
            j = mid;
        } else {
            i = mid;
        }
    }

    const value = getWireValueCaption(kind, wave[i][1], wave[i][2]);
    return value;
}


export const scopes = new Set(['begin', 'fork', 'function', 'module', 'task']);
export const variables = new Set(['event', 'integer', 'parameter', 'real', 'realtime', 'reg', 'supply0',
                           'supply1', 'time', 'tri', 'triand', 'trior', 'trireg', 'tri0', 'tri1',
                           'wand', 'wire', 'wor', 'string']);

export function isScope(wire) {
    return scopes.has(wire.type);
}

export function isVariable(wire) {
    return variables.has(wire.type);
}

export function webglColor2rgba(r, g, b, a) {
    const standRed = Math.floor(r * 255);
    const standGreen = Math.floor(g * 255);
    const standBlue = Math.floor(b * 255);
    return `rgba(${standRed},${standGreen},${standBlue},${a})`;
}

/**
 * 
 * @param {string} rgba 
 * @returns {{
 *  red: number,
 *  green: number,
 *  blue: number,
 *  alpha: number
 * } | undefined}
 */
export function rgba2WebglColor(rgba) {
    const match = rgba.match(/^rgba\((\d+),\s*(\d+),\s*(\d+),\s*(\d+(\.\d+)?)\)$/);  
    if (match) {  
        return {  
            red: parseInt(match[1]) / 255,  
            green: parseInt(match[2]) / 255,
            blue: parseInt(match[3]) / 255,  
            alpha: parseFloat(match[4])
        };  
    } else {  
        return undefined;
    }  
}

export const predefinedColors = [
    '#ff4500',
    '#ff8c00',
    '#ffd700',
    '#90ee90',
    '#00ced1',
    '#1e90ff',
    '#c71585',
    'rgba(255, 69, 0, 0.68)',
    'rgb(255, 120, 0)',
    'hsv(51, 100, 98)',
    'hsva(120, 40, 94, 0.5)',
    'hsl(181, 100%, 37%)',
    'hsla(209, 100%, 56%, 0.73)',
    '#c7158577'
];