
/**
 * @description
 * @typedef {Object} VecRenderNumberVertices
 * @property {number[]} lineVertices
 * @property {number[]} maskVertices
 */

/**
 * @description
 * @typedef {Object} WaveValueFormatMaxMin
 * @property {number} maxVal
 * @property {number} minVal
 */

/**
 * @description
 * @typedef {Object} BitRenderTempStruct
 * @property {number} color 颜色的索引 颜色rgb = gl_Colors[color]
 * @property {number} y 裁剪空间的纵坐标
 */

/**
 * @description
 * @typedef {Object} BitRenderHandlePoint
 * @property {number} x
 * @property {number} y
 * @property {number} color
 */

import { globalLookup } from "@/hook/global";
import { lineAnalog_WidthShift, lineAnlog_GL_WidthShifts, maskColorIndexOffset, posYFactor, prettyPrint } from "../render-utils";
import { FormatValueRender, hexToSignedInt, JSValueRender } from "./renderFormat";

function makeQuadVertices(p1, p2, p3, p4) {
    return [
        ...p1, ...p2, ...p3,
        ...p1, ...p3, ...p4
    ]
}

/**
 * @description 根据四个点计算有顺序的 triangle 类型的，其中，p1p2组成单独的 frag，p3p4组成单独的frag
 * @param {number[]} p1 
 * @param {number[]} p2 
 * @param {number[]} p3 
 * @param {number[]} p4 
 * @returns 
 */
function makeOrderedQuadVertices(p1, p2, p3, p4) {
    let t;
    if (p1[1] < p2[1]) {
        t = p1;
        p1 = p2;
        p2 = t;        
    }

    if (p3[1] < p4[1]) {
        t = p3;
        p3 = p4;
        p4 = t;
    }

    // 保证 p1 的 y 比 p2 大
    // 保证 p3 的 y 比 p4 大
    return [
        ...p1, ...p3, ...p2,
        ...p2, ...p3, ...p4
    ]
}

/**
 * 
 * @param {number} x1 
 * @param {number} y1 
 * @param {number} x2 
 * @param {number} y2 
 * @param {number} color 
 * @param {number} wsIndex
 * @returns {number[]} 
 */
function makeRectangleVertices(x1, y1, x2, y2, color, wsIndex = 0) {
    // uvec2 pos; uvec3 control;
    const r1 = [x1, y2 * posYFactor, 0, wsIndex, color];
    const r2 = [x1, y1 * posYFactor, 0, wsIndex, color];
    const r3 = [x2, y1 * posYFactor, 0, wsIndex, color];
    const r4 = [x2, y2 * posYFactor, 0, wsIndex, color];
    return makeQuadVertices(r1, r2, r3, r4);
}

/**
 * @description 将 bit 类型的 wave 的值转化为渲染需要用到的值
 * @param {string} link
 * @param {number[]} wave 
 * @param {number} time 
 * @return {BitRenderTempStruct}
*/
function makeBitRenderParam(link, wave, time) {
    const value = wave[1];
    let colorParam;
    switch (value) {
        case 0:             colorParam = { y: -1, color: 2 }; break;       // 0 value
        case 1:             colorParam = { y: 1,  color: 3 }; break;       // 1 value
        case 2: case 3:     colorParam = { y: -1, color: 4 }; break;       // 不定态 x
        case 4: case 5:     colorParam = { y: 0,  color: 2 }; break;       // 高阻态 z
        default:            colorParam = { y: -1, color: 7 }; break;       // 其他，我也不知道还有啥
    }

    if (value === 0 || value === 1) {
        const renderOptions = globalLookup.currentSignalRenderOptions;
        if (renderOptions.has(link)) {
            const option = renderOptions.get(link);
            if (typeof option.color === 'number') {
                colorParam.color = option.color;
            }
        }
    }

    return colorParam;
}


/**
 * 
 * @param {BitRenderHandlePoint | undefined} p0 前一个点
 * @param {BitRenderHandlePoint | undefined} p1 当前的点
 * @param {BitRenderHandlePoint | undefined} p2 后一个点
 * @returns {number} 这是 widthshift 的索引，只需要 + 4 再 % 8 就能得到另一个 
 */
function makeWidthShiftIndexByPoints(p0, p1, p2) {
    if (p0 === undefined) {
        if (p1.y === p2.y) {
            return 0;
        } else if (p1.x === p2.x) {
            return 6;
        }
    } else if (p2 === undefined) {
        if (p1.y === p0.y) {
            return 0;
        } else if (p1.x === p0.x) {
            return 6;
        }
    } else {
        if (p0.x !== p1.x && p0.y === p1.y && p1.x === p2.x && p1.y !== p2.y) {
            if (p2.y > p1.y) {
                return 1;
            } else {
                return 7;
            }
        } else if (p0.x === p1.x && p0.y !== p1.y && p1.x !== p2.x && p1.y === p2.y) {
            if (p1.y > p0.y) {
                return 1;
            } else {
                return 7;
            }
        }
    }

    return 0;
}


/**
 * 
 * @param {string} link
 * @param {Array<string | number>} wave 
 * @param { number } time 
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsBit(link, wave, time) {
    const length = wave.length;
    // 先将节点数据转化为裁剪空间的坐标
    const perspectivePoints = [];

    for (let i = 0; i < length; ++ i) {
        // const currentWave = wave[(i === 0) ? 0 : (i - 1)];
        const currentWave = wave[i];
        const nextWave = (i === (length - 1)) ? wave[i] : wave[i + 1];

        const t1 = currentWave[0];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
        const value1 = currentWave[1];
        const value2 = nextWave[1];

        const renderParam1 = makeBitRenderParam(link, currentWave, time);
        const renderParam2 = makeBitRenderParam(link, nextWave, time);

        if (i === 0) {
            perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
            perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
        } else {
            const lastPoint = perspectivePoints[perspectivePoints.length - 1];
            if ((lastPoint.y !== renderParam1.y) || (lastPoint.color !== renderParam1.color)) {
                perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
            }
            if ((renderParam1.y !== renderParam2.y) || (renderParam1.color !== renderParam2.color)) {
                perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
            }
        }
    }

    // 确保最后一个点延申到了 time
    const lastPoint = perspectivePoints[perspectivePoints.length - 1];
    if (lastPoint.x < time) {
        perspectivePoints.push({ x: time, y: lastPoint.y, color: lastPoint.color });
    }

    // 计算出传入 shader 四元组数组
    // 四元组： (x, yshift_index, color_index, width_shift_index)
    const pointNum = perspectivePoints.length;
    const lineVertices = [];
    const maskVertices = [];

    
    // 制作 lineVertices
    for (let i = 0; i < pointNum; ++ i) {
        // p0: 上一个点
        // p1: 当前的点
        // p2: 下一个点
        const p0 = perspectivePoints[i - 1];
        const p1 = perspectivePoints[i];
        const p2 = perspectivePoints[i + 1];

        const wsIndex = makeWidthShiftIndexByPoints(p0, p1, p2);

        if (wsIndex === undefined) {
            console.log(p0, p1, p2);
        }

        lineVertices.push(
            // uvec2 pos; uvec3 control;
            p1.x, p1.y * posYFactor, 0, wsIndex,           p1.color,
            p1.x, p1.y * posYFactor, 0, (wsIndex + 4) % 8, p1.color
        );

        // 防止颜色不同导致单个图元内出现两个颜色，这会引发shader的渐变
        if (p2 !== undefined && p1.color !== p2.color) {
            lineVertices.push(
                // uvec2 pos; uvec3 control;
                p1.x, p1.y * posYFactor, 0, wsIndex,           p2.color,
                p1.x, p1.y * posYFactor, 0, (wsIndex + 4) % 8, p2.color,
            );
        }
    }

    // 制作 maskVertices
    for (let i = 0; i < pointNum; ++ i) {
        const p1 = perspectivePoints[i];

        // 开头就有一条线
        if (i === 0 && p1.y === 1) {
            while (perspectivePoints[++ i] && perspectivePoints[i].y === 1);
            // 回退
            if (-- i > 0) {
                // 四元组： (x, yshift_index, color_index, width_shift_index)
                const rectangleVertices = makeRectangleVertices(p1.x, 1, perspectivePoints[i].x, -1, p1.color + maskColorIndexOffset, 4);
                // 三角图元画矩形
                maskVertices.push(...rectangleVertices);
                continue;
            }
        }

        // 上升沿才使用
        const p2 = perspectivePoints[i + 1];
        const p3 = perspectivePoints[i + 2];

        if (p1 === undefined || p2 === undefined || p3 === undefined) {
            continue;
        }
        if (p2.y > p1.y) {
            // 矩形的四个点
            // 四元组： (x, yshift_index, color_index, width_shift_index)
            const rectangleVertices = makeRectangleVertices(p2.x, p2.y, p3.x, p1.y, p2.color + maskColorIndexOffset, 4);
            // 三角图元画矩形
            maskVertices.push(...rectangleVertices);
        }
    }

    return { lineVertices, maskVertices };
}

/**
 * @description Digital
 * @param {GlobalLookup} lookup
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsCommonDigital(lookup, link, wave, time) {
    const length = wave.length;
    const lineVertices = [];
    const maskVertices = [];

    // 获取外部自定义颜色
    // 这部分颜色只会染色非异常部分的信号
    let userDefineCommonColor = getUserDefineCommonColor(lookup, link);
    
    for (let i = 0; i < length; ++ i) {
        const [t1, val, mask] = wave[i];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];

        // t1(val) --- t2(val)
        // 详见 设计图 ./design/webgl.drawio makeVecVertex原理 sheet
        // 详见 gl_Shifts_for_bar 的设计
        const p0 = {x: t1, y: 0, shift: 0};
        const p1 = {x: t2, y: 0, shift: 0};
        const a0 = {x: t1, y: -1, shift: 4};
        const a1 = {x: t2, y: -1, shift: 5};
        const a2 = {x: t1, y: 1, shift: 1};
        const a3 = {x: t2, y: 1, shift: 2};

        // 选择当前的颜色
        // 颜色的定义在 gl_Colors_template 中
        // 默认为 5，vec 专用颜色
        let color = 5;
        const renderOptions = lookup.currentSignalRenderOptions;
        if (mask) {
            if (val) {
                // ? 代表 unknown，比如 b1xx0xx0xx1xx000xxxx0 这种，一部分高阻的就是
                color = 7;
            } else {
                // 颜色 4 代表高阻态
                color = 4;
            }
        } else if (userDefineCommonColor !== undefined) {
            color = userDefineCommonColor;
        }

        const points = [ a1, p1, a3, a2, p0, a0 ];
        const wsIndice = [ 1, 2, 3, 5, 6, 7 ];
        for (let j = 0; j < 6; ++ j) {
            const prePoint = points[j];
            const nextPoint = points[(j + 1) % 6];
            const preWsIndex = wsIndice[j];
            const nextWsIndex = wsIndice[(j + 1) % 6];
            const quadVertices = makeQuadVertices( 
                [prePoint.x,  prePoint.y * posYFactor, prePoint.shift, preWsIndex, color],
                [prePoint.x,  prePoint.y * posYFactor, prePoint.shift, (preWsIndex + 4) % 8, color],
                [nextPoint.x, nextPoint.y * posYFactor, nextPoint.shift, (nextWsIndex + 4) % 8, color],
                [nextPoint.x, nextPoint.y * posYFactor, nextPoint.shift, nextWsIndex, color],
            );
            lineVertices.push(...quadVertices);
        }

        const maskTriangles = [
            [3, 4, 5],
            [5, 0, 3],
            [0, 3, 2],
            [0, 1, 2]
        ];
        
        for (const triangleIndice of maskTriangles) {
            const [i1, i2, i3] = triangleIndice;
            const point1 = points[i1];
            const point2 = points[i2];
            const point3 = points[i3];
            maskVertices.push(
                point1.x, point1.y * posYFactor, point1.shift, wsIndice[i1], color + maskColorIndexOffset,
                point2.x, point2.y * posYFactor, point2.shift, wsIndice[i2], color + maskColorIndexOffset,
                point3.x, point3.y * posYFactor, point3.shift, wsIndice[i3], color + maskColorIndexOffset,
            );
        }
    }

    return { lineVertices, maskVertices };
}


/**
 * 
 * @param {GlobalLookup} lookup 
 * @param {string} link 
 */
function getValFormatCode(lookup, link) {
    const renderOptions = lookup.currentSignalRenderOptions;
    if (renderOptions.has(link)) {
        const option = renderOptions.get(link);
        if (typeof option.valueFormat === 'number') {
            return option.valueFormat;
        } else {
            return 0;
        }
    }

    return 0;
}

/**
 * @description 计算为 js 的数字，如果当前模式为 有符号整型，需要先转换成 有符号数进行比较
 * @param {number} formatCode 
 * @param {number | string} val 
 * @param {number} width 
 */
function explainAsJSNumber(formatCode, val, width) {
    if (formatCode === 3) {
        if (typeof val === 'number') {
            val = val.toString(16);
        }
        if (val.startsWith('0x')) {
            val = val.substring(2);
        }
        if (val.length < width) {
            val = '0'.repeat(width - val.length) + val;
        }
        return hexToSignedInt(val);
    } else {
        return parseInt(val);
    }
}

/**
 * @description Analog (Ladder) 阶梯状
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @param {number} formatCode
 * @param {number} width
 * @returns {WaveValueFormatMaxMin} 
 */
function getMaxMinByFormat(link, wave, time, formatCode, width) {
    let maxVal = undefined;
    let minVal = undefined;
    const length = wave.length;

    const valueRender = new JSValueRender(link, width);

    // TODO : 优化这段代码
    for (let i = 0; i < length; ++ i) {
        const [t1, val, mask] = wave[i];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
        const numVal = valueRender.explainAsNumber(val);

        if (maxVal === undefined) {
            maxVal = numVal;
        } else {
            maxVal = Math.max(maxVal, numVal);
        }

        if (minVal === undefined) {
            minVal = numVal;
        } else {
            minVal = Math.min(minVal, numVal);
        }
    }

    return { maxVal, minVal };
}

/**
 * 
 * @param {number} formatCode 
 * @param {number} maxVal 
 * @param {number} minVal 
 * @returns {(val: number) => number}
 */
function getMappingFunc(formatCode, maxVal, minVal) {
    if (formatCode === 3) {
        const maxAbs = Math.max(Math.abs(maxVal), Math.abs(minVal));
        if (maxAbs === 0) {
            return val => 0;
        }
        const k = 1 / maxAbs;
        const b = -1;
        // 把 [-maxAbs, +maxAbs] 映射到 [-1, 1]
        return val => k * (val + maxAbs) + b;
    } else {
        if (maxVal === 0) {
            return val => -1;
        }
        const k = 2 / maxVal;
        const b = -1;
        // 把 [0, max] 映射到 [-1, 1]
        return val => k * val + b;
    }
}

/**
 * @description Analog (Ladder) 阶梯状
 * @param {GlobalLookup} lookup
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsLadderAnalog(lookup, link, wave, time) {
    const formatCode = getValFormatCode(lookup, link);
    const signal = lookup.link2CurrentWires.get(link);
    const width = signal.size;

    const { maxVal, minVal } = getMaxMinByFormat(link, wave, time, formatCode, width);
    const coordinateTransform = getMappingFunc(formatCode, maxVal, minVal);

    // 数值解释器，负责将对应的数值翻译成渲染用的 y
    const valueRender = new JSValueRender(link, width);

    // 获取外部自定义颜色
    // 这部分颜色只会染色非异常部分的信号
    let userDefineCommonColor = getUserDefineCommonColor(lookup, link);

    function makeLadderAnalogRenderParam(link, wave, time) {        
        const [t1, val, mask] = wave;

        if (mask) {
            // 颜色 4 代表高阻态
            // 7 代表 unknown，比如 b1xx0xx0xx1xx000xxxx0 这种，一部分高阻的就是
            let color = val ? 7 : 4;

            return { y: -1, color };  
        }

        // 根据当前格式进行转换
        const numVal = valueRender.explainAsNumber(val);
        // 此时的 y 是一个 -1 到 1 的浮点数，因为 VAO 我设置了 Int
        const y = coordinateTransform(numVal);
        const colorParam = { y, color: 5 };

        if (userDefineCommonColor !== undefined) {
            colorParam.color = userDefineCommonColor;
        }

        return colorParam;
    }

    const length = wave.length;
    // 先将节点数据转化为裁剪空间的坐标
    const perspectivePoints = [];

    for (let i = 0; i < length; ++ i) {
        // const currentWave = wave[(i === 0) ? 0 : (i - 1)];
        const currentWave = wave[i];
        const nextWave = (i === (length - 1)) ? wave[i] : wave[i + 1];

        const t1 = currentWave[0];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
        const value1 = currentWave[1];
        const value2 = nextWave[1];

        const renderParam1 = makeLadderAnalogRenderParam(link, currentWave, time);
        const renderParam2 = makeLadderAnalogRenderParam(link, nextWave, time);

        if (i === 0) {
            perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
            perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
        } else {
            const lastPoint = perspectivePoints[perspectivePoints.length - 1];
            if ((lastPoint.y !== renderParam1.y) || (lastPoint.color !== renderParam1.color)) {
                perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
            }
            if ((renderParam1.y !== renderParam2.y) || (renderParam1.color !== renderParam2.color)) {
                perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
            }
        }
    }

    // 确保最后一个点延申到了 time
    const lastPoint = perspectivePoints[perspectivePoints.length - 1];
    if (lastPoint.x < time) {
        perspectivePoints.push({ x: time, y: lastPoint.y, color: lastPoint.color });
    }

    // 计算出传入 shader 四元组数组
    // 四元组： (x, yshift_index, color_index, width_shift_index)
    const pointNum = perspectivePoints.length;
    const lineVertices = [];
    const maskVertices = [];

    // 制作 lineVertices
    for (let i = 0; i < pointNum; ++ i) {
        // p0: 上一个点
        // p1: 当前的点
        // p2: 下一个点
        const p0 = perspectivePoints[i - 1];
        const p1 = perspectivePoints[i];
        const p2 = perspectivePoints[i + 1];

        const wsIndex = makeWidthShiftIndexByPoints(p0, p1, p2);

        lineVertices.push(
            // uvec2 pos; uvec3 control;
            p1.x, p1.y * posYFactor, 0, wsIndex,           p1.color,
            p1.x, p1.y * posYFactor, 0, (wsIndex + 4) % 8, p1.color
        );

        // 防止颜色不同导致单个图元内出现两个颜色，这会引发shader的渐变
        if (p2 !== undefined && p1.color !== p2.color) {
            lineVertices.push(
                // uvec2 pos; uvec3 control;
                p1.x, p1.y * posYFactor, 0, wsIndex,           p2.color,
                p1.x, p1.y * posYFactor, 0, (wsIndex + 4) % 8, p2.color,
            );
        }
    }

    // 制作 maskVertices
    for (let i = 0; i < pointNum; ++ i) {
        const p1 = perspectivePoints[i];

        // 开头就有一条线
        if (i === 0 && p1.y === 1) {
            while (perspectivePoints[++ i] && perspectivePoints[i].y === 1);
            // 回退
            if (-- i > 0) {
                // 四元组： (x, yshift_index, color_index, width_shift_index)
                const rectangleVertices = makeRectangleVertices(p1.x, 1, perspectivePoints[i].x, -1, p1.color + maskColorIndexOffset, 4);
                // 三角图元画矩形
                maskVertices.push(...rectangleVertices);
                continue;
            }
        }

        // 上升沿才使用
        const p2 = perspectivePoints[i + 1];
        const p3 = perspectivePoints[i + 2];

        if (p1 === undefined || p2 === undefined || p3 === undefined) {
            continue;
        }
        if (p2.y > -1) {
            // 矩形的四个点
            const rectangleVertices = makeRectangleVertices(p2.x, p2.y, p3.x, -1, p2.color + maskColorIndexOffset, 4);
            // 三角图元画矩形
            maskVertices.push(...rectangleVertices);
        }
    }

    // prettyPrint(lineVertices);

    return { lineVertices, maskVertices };
}



/**
 * 
 * @param {number[]} p1 
 * @param {number[]} p2 
 * @param {number[]} p3
 * @returns {number} 
 */
function getSlope(p1, p2, p3) {
    if (p3 === undefined) {

    } else {

    }
}


/**
 * @description 获取用户定义这个信号的颜色，只选择正常部分的颜色
 * @param {GlobalLookup} lookup
 * @param {string} link 
 * @returns {number | undefined}
 */
function getUserDefineCommonColor(lookup, link) {
    const renderOptions = lookup.currentSignalRenderOptions;
    if (renderOptions.has(link)) {
        const option = renderOptions.get(link);
        if (typeof option.color === 'number') {
            return option.color;
        }
    }
    return undefined;
}

/**
 * @description Analog (Line) 模拟状
 * @param {GlobalLookup} lookup
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsLineAnalog(lookup, link, wave, time) {
    const formatCode = getValFormatCode(lookup, link);
    const signal = lookup.link2CurrentWires.get(link);
    const width = signal.size;

    const { maxVal, minVal } = getMaxMinByFormat(link, wave, time, formatCode, width);
    const coordinateTransform = getMappingFunc(formatCode, maxVal, minVal);

    const length = wave.length;
    const lineVertices = [];
    const maskVertices = [];

    // 数值解释器，负责将对应的数值翻译成渲染用的 y
    const valueRender = new JSValueRender(link, width);

    // 获取外部自定义颜色
    // 这部分颜色只会染色非异常部分的信号
    let userDefineCommonColor = getUserDefineCommonColor(lookup, link);

    function makeLineAnalogRenderParam(link, wave, time) {
        const [t1, val, mask] = wave;

        if (mask) {
            // 颜色 4 代表高阻态
            // 7 代表 unknown，比如 b1xx0xx0xx1xx000xxxx0 这种，一部分高阻的就是
            let color = val ? 7 : 4;

            // 不定态
            return { y: -1, color };  
        }

        // 根据当前格式进行转换
        const numVal = valueRender.explainAsNumber(val);
        // 此时的 y 是一个 -1 到 1 的浮点数，因为 VAO 我设置了 Int
        const y = coordinateTransform(numVal);
        const colorParam = { y, color: 5 };

        if (userDefineCommonColor !== undefined) {
            colorParam.color = userDefineCommonColor;
        }

        return colorParam;
    }

    for (let i = 0; i < length; ++ i) {
        // const currentWave = wave[(i === 0) ? 0 : (i - 1)];
        const currentWave = wave[i];
        const nextWave = (i === (length - 1)) ? wave[i] : wave[i + 1];

        const t1 = currentWave[0];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];

        const currentParam = makeLineAnalogRenderParam(link, currentWave, time);
        const nextParam = makeLineAnalogRenderParam(link, nextWave, time);
        
        const color = currentParam.color;
        const maskColor = color + maskColorIndexOffset;
        const y1 = currentParam.y;
        const y2 = nextParam.y;
        // 给每一个点构造正方形
        // 需要渲染的部分为 current 的正方形和 current 正方形和 next 正方形中间的连线
        
        lineVertices.push(...makeQuadVertices(
            [t1, y1 * posYFactor, 0, 1, color],
            [t1, y1 * posYFactor, 0, 3, color],
            [t1, y1 * posYFactor, 0, 5, color],
            [t1, y1 * posYFactor, 0, 7, color],
        ));

        // 根据 p1.y 和 p2.y 的相对位置，一共分为三种情况
        if (y1 > y2) {
            // y1 > y2
            lineVertices.push(...makeQuadVertices(
                [t1, y1 * posYFactor, 0, 3, color],
                [t1, y1 * posYFactor, 0, 7, color],
                [t2, y2 * posYFactor, 0, 7, color],
                [t2, y2 * posYFactor, 0, 3, color],
            ));

        } else if (y1 === y2) {
            // y1 = y2
            lineVertices.push(...makeQuadVertices(
                [t1, y1 * posYFactor, 0, 5, color],
                [t1, y1 * posYFactor, 0, 7, color],
                [t2, y2 * posYFactor, 0, 1, color],
                [t2, y2 * posYFactor, 0, 3, color],
            ));
        } else {
            // y1 < y2
            lineVertices.push(...makeQuadVertices(
                [t1, y1 * posYFactor, 0, 5, color],
                [t1, y1 * posYFactor, 0, 1, color],
                [t2, y2 * posYFactor, 0, 1, color],
                [t2, y2 * posYFactor, 0, 5, color],
            ));
        }
    }

        // if (y1 > -1 && y2 > -1) {
        //     maskVertices.push(...makeQuadVertices(
        //         [t1, factY1,       0, 0, maskColor],
        //         [t2, factY2,       0, 0, maskColor],
        //         [t2, - posYFactor, 0, 0, maskColor],
        //         [t1, - posYFactor, 0, 0, maskColor],
        //     ));
        // } else if (y1 > -1 && y2 === -1) {
        //     maskVertices.push(
        //         t1, factY1,       0, 0, maskColor,
        //         t2, - posYFactor, 0, 0, maskColor,
        //         t1, - posYFactor, 0, 0, maskColor,
        //     )
        // } else if (y1 === -1 && y2 > -1) {
        //     maskVertices.push(
        //         t1, - posYFactor, 0, 0, maskColor,
        //         t2, factY2,       0, 0, maskColor,
        //         t2, - posYFactor, 0, 0, maskColor,
        //     );
        // }
    return { lineVertices, maskVertices };
}